import pytest
import numpy as np

from cocoex import BareProblem
from cocoex.exceptions import NoSuchSuiteException


def test_bad_suite():
    with pytest.raises(NoSuchSuiteException):
        BareProblem("bad_suite", 1, 1, 1)


def test_bbob():
    for fid in range(1, 25):
        for dimension in [2, 3, 4, 5, 7, 11, 13, 17, 18, 19, 20, 31, 40]:
            x0 = np.zeros(dimension)
            for instance in [0, 1, 10, 100, 1000]:
                fn = BareProblem("bbob", fid, dimension, instance)
                assert str(fn) == f"bbob_f{fid:03d}_i{instance:02d}_d{dimension:02d}"
                assert fn(x0) >= fn.best_value()
                assert not fn.is_tainted
                assert np.isclose(fn.best_value(), fn(fn.best_parameter()))
                assert fn.is_tainted


def test_list():
    fn = BareProblem("bbob", 1, 4, 1)
    assert fn([1, 2, 3, 4]) >= fn([1.0, 2.0, 3.0, 4.0])


def test_multiple_parameters():
    for n in [1, 2, 10, 100, 200]:
        fn = BareProblem("bbob", 1, 4, 1)
        X = np.random.uniform(-5, 5, size=(n, 4))
        y = fn(X)
        assert len(y) == n
        assert np.all(y >= fn.best_value())



"""Experimentation module of the COCO - COmparing Continuous Optimizers -
framework.

The module provides benchmark test beds in the `Suite` class
and output data facilities in the `Observer` class.

See the documentation of the `Suite` class:

>>> import cocoex as ex
>>> help(ex.Suite)  # doctest: +ELLIPSIS
Help on class Suite...
>>> print(ex.known_suite_names)  # doctest: +ELLIPSIS
[...

A more complete example use case can be found in the `example_experiment.py`
file.
"""
from . import solvers # noqa: F401
from . import utilities # noqa: F401
from . import exceptions # noqa: F401
from .interface import Observer as _Observer
from .interface import Problem as _Problem
from .interface import Suite as _Suite 
from .interface import known_suite_names
from .utilities import ExperimentRepeater
from .utilities import BatchScheduler
from .bare_problem import BareProblem
from .interface import log_level # noqa: F401
from ._version import __version__ # noqa: F401


__all__ = [
    'Observer',
    'Suite',
    'known_suite_names', 
    'ExperimentRepeater', 
    'BatchScheduler',
    'default_observers',
    'BareProblem'
]


def default_observers(update=None):
    """return a map from suite names to default observer names.

    This function can also be used to update this map using
    a `dict` or a `list` of key-value pairs.
    """
    # this is a function only to make the doc available and
    # because @property doesn't work on module level
    _default_observers.update(update or {})
    return _default_observers
_default_observers = {
    'bbob': 'bbob',
    'bbob-biobj': 'bbob-biobj',
    'bbob-biobj-ext': 'bbob-biobj',
    'bbob-constrained': 'bbob',
    'bbob-largescale': 'bbob',
    'bbob-mixint': 'bbob',
    'bbob-biobj-mixint': 'bbob-biobj',
    }

class Suite(_Suite):
    """Suite of benchmark problems.

    Input arguments to `Suite` are `name: str`, `instance: str`, `options: str`,
    and passed to the respective C code (see `coco.h`).

    >>> import cocoex as ex
    >>> suite = ex.Suite("bbob", "", "")
    >>> f = suite.next_problem()
    >>> assert f.number_of_objectives == 1
    >>> assert f.evaluations == 0
    >>> print("f([1,2]) = %.11f" % f([1,2]))
    f([1,2]) = 90.00369408000
    >>> assert f.evaluations == 1

    Sweeping through all problems is as simple as:

    >>> import cocoex as ex
    >>> suite = ex.Suite("bbob-biobj", "", "")
    >>> observer = ex.Observer("bbob-biobj", "result_folder:doctest")
    >>> for fun in suite:
    ...     if fun.index == 0:
    ...         print("Number of objectives %d, %d, %d" %
    ...                 (fun.number_of_objectives,
    ...                  suite.number_of_objectives[0],
    ...                  suite.number_of_objectives[-1]))
    ...     fun.observe_with(observer)
    ...     assert fun.evaluations == 0
    ...     assert fun.number_of_objectives == suite.number_of_objectives[0]
    ...     # run run run using fun  # doctest: +ELLIPSIS
    Number of objectives 2, 2, 2...

    In the example, an observer was added to produce output data for the
    COCO post-processing.

    The following example runs the entire bbob2009 benchmark suite
    on random search:

    >>> import numpy as np
    >>> from cocoex import Suite, Observer
    ...
    >>> MAX_FE = 22  # max f-evaluations
    >>> def random_search(f, lb, ub, m):  # don't use m >> 1e5 with this implementation
    ...     candidates = lb + (ub - lb) * np.random.rand(m, len(lb))
    ...     return candidates[np.argmin([f(x) for x in candidates])]
    ...
    >>> solver = random_search
    >>> suite = Suite("bbob", "year:2009", "")
    >>> observer = Observer("bbob",
    ...              "result_folder: %s_on_%s" % (solver.__name__, "bbob2009"))
    >>> for fun in suite:
    ...     assert fun.evaluations == 0
    ...     if fun.dimension >= 10:
    ...         break
    ...     print('Current problem index = %d' % fun.index)
    ...     fun.observe_with(observer)
    ...     assert fun.evaluations == 0
    ...     solver(fun, fun.lower_bounds, fun.upper_bounds, MAX_FE)
    ...     # data should be now in the "exdata/random_search_on_bbob2009" folder
    ...     assert fun.evaluations == MAX_FE  # depends on the solver
    ...     # doctest: +ELLIPSIS
    Current problem index = 0...
    >>> #
    >>> # Exactly the same using another looping technique:
    >>> for id in suite.ids():
    ...     fun = suite.get_problem(id, observer)
    ...     _ = solver(fun, fun.lower_bounds, fun.upper_bounds, MAX_FE)
    ...     print("Evaluations on %s: %d" % (fun.name, fun.evaluations))
    ...     fun.free()  # this is absolutely necessary here
    ...     # doctest: +ELLIPSIS
    Evaluations on ...

    We can select a single function, say BBOB f9 in 20D, of a given suite like:

    >>> import cocoex as ex
    >>> suite = ex.Suite("bbob", "", "dimensions:20 instance_indices:1")
    >>> len(suite)
    24
    >>> f9 = suite.get_problem(8)
    >>> x = f9.initial_solution  # a copy of a feasible point
    >>> all(x == 0)
    True

    See module attribute `cocoex.known_suite_names` for known suite names:

    >>> import cocoex as ex
    >>> for suite_name in ex.known_suite_names:
    ...     suite = ex.Suite(suite_name, "", "")
    ...     print(suite.dimensions)
    ...     for f in suite:
    ...         assert f.dimension in suite.dimensions
    ...         assert f.evaluations == 0
    ...         # doctest: +ELLIPSIS
    [2, 3, 5, 10, 20, 40]...

    See file `example_experiment.py` for a full example use case.

    Details: depending on the benchmark suite and observer, only one problem can
    be open at a time. Using `get_problem` without `free` or mixing the use of
    `next_problem` and `get_problem` may not be possible. For example, in this
    case the "bbob" observer is known to lead to a crash of the Python
    interpreter.

    See also `Observer` and `example_experiment.py`.
    """
    def __init__(self, suite_name, suite_instance, suite_options):
        """``suite_instance`` and ``suite_options`` can be empty strings."""
        # this __init__ defines the arguments for _Suite.__cinit__,
        # which is called implicitly. Calling the super class init fails in Python 3.
        # super(Suite, self).__init__(suite_name, suite_instance, suite_options)
        # _Suite.__cinit__(self, suite_name, suite_instance, suite_options)
    def reset(self):
        """reset to original state, affecting `next_problem()`,
        `current_problem`, `current_index`"""
        super().reset()
    def next_problem(self, observer=None):
        """return the "next" problem in this `Suite`.

        return the first problem on the first call or after
        `reset` ().

        `next_problem` serves to sweep through the `Suite` smoothly.
        """
        return super().next_problem(observer)
    def get_problem(self, id, observer=None):
        """return a `Problem` instance, by default unobserved, using ``id: str``
        or index (where ``id: int``) to identify the desired problem.

        All values between zero and `len(self) - 1` are valid index values:

        >>> import cocoex as ex
        >>> suite = ex.Suite("bbob-biobj", "", "")
        >>> for index in range(len(suite)):
        ...     problem = suite.get_problem(index)
        ...     # work work work using problem
        ...     problem.free()

        A shortcut for `suite.get_problem(index)` is `suite[index]`, they are
        synonym.

        Details:
            - Here an `index` takes values between 0 and `len(self) - 1` and can in
              principle be different from the problem index in the benchmark suite.
    
            - This call does not affect the state of the `current_problem` and
              `current_index` attributes.
    
            - For some suites and/or observers, the `free()` method of the problem
              must be called before the next call of `get_problem`. Otherwise Python
              might just silently die, which is e.g. a known issue of the "bbob"
              observer.

        See also `ids`.
        """
        return super().get_problem(id, observer)

    def get_problem_by_function_dimension_instance(self, function, dimension, instance, observer=None):
        """return a `Problem` instance, by default unobserved, using function,
        dimension and instance to identify the desired problem.

        If a suite contains multiple problems with the same function, dimension
        and instance, the first corresponding problem is returned.

        >>> import cocoex as ex
        >>> suite = ex.Suite("bbob-biobj", "", "")
        >>> problem = suite.get_problem_by_function_dimension_instance(1, 2, 3)
        >>> # work work work using problem
        >>> problem.free()

        Details:
            - Function, dimension and instance are integer values from 1 on.
    
            - This call does not affect the state of the `current_problem` and
              `current_index` attributes.
    
            - For some suites and/or observers, the `free()` method of the problem
              must be called before the next call of
              `get_problem_by_function_dimension_instance`. Otherwise Python might
              just silently die, which is e.g. a known issue of the "bbob" observer.
        """
        return super().get_problem_by_function_dimension_instance(
                function, dimension, instance, observer)

    def __getitem__(self, key):
        """`self[i]` is a synonym for `self.get_problem(i)`, see `get_problem`
        """
        return self.get_problem(key)

    def free(self):
        """free underlying C structures"""
        super().free()

    def find_problem_ids(self, *args, **kwargs):
        """has been renamed to `ids`"""
        raise NotImplementedError(
            "`find_problem_ids()` has been renamed to `ids()`")

    def ids(self, *id_snippets, **kwargs):  # get_problem=False, verbose=False):
        """`ids(*id_snippets, get_problem=False, verbose=False)`
        return all problem IDs that contain all of the `id_snippets`.

        An ID can be used for indexing, that is, when calling the method
        `get_problem(id)`.

        If `get_problem is True`, the problem for the first matching ID is
        returned.

        >>> import cocoex as ex
        >>> s = ex.Suite("bbob", "", "")
        >>> s.ids("f001", "d10", "i01")
        ['bbob_f001_i01_d10']

        We can sweep through all instances of the ellipsoidal function f10
        in 20-D of the BBOB suite like this:

        >>> import cocoex as ex
        >>> suite = ex.Suite("bbob", "", "")
        >>> ids = suite.ids("f010", "d20")
        >>> used_indices = []
        >>> for p in suite:
        ...     if p.id in ids:
        ...         # work work work with problem `p`
        ...         used_indices.append(p.index)
        >>> print(used_indices)
        [1575, 1576, 1577, 1578, 1579, 1580, 1581, 1582, 1583, 1584, 1585, 1586, 1587, 1588, 1589]

        A desired problem can also be filtered out during creation:

        >>> import cocoex as ex
        >>> f9 = ex.Suite("bbob", "",
        ...               "function_indices:9 dimensions:20 instance_indices:1-5")[0]
        >>> print(f9.id)
        bbob_f009_i01_d20

        """
        return super().ids(*id_snippets, **kwargs)

    @property
    def current_problem(self):
        """current "open/active" problem to be benchmarked"""
        return super().current_problem
    @property
    def current_index(self):
        """index in the enumerator of all problems in this suite.

        Details: To get the index in the underlying C implementation, which
        usually matches `current_index` one-to-one, use:

        >>> import cocoex as ex
        >>> suite = ex.Suite("bbob", "", "")
        >>> suite.current_index is None
        True
        >>> suite.next_problem().id[-17:].lower()
        'bbob_f001_i01_d02'
        >>> suite.current_index, suite.indices[suite.current_index]
        (0, 0)

        """
        return super().current_index
    @property
    def problem_names(self):
        """list of problem names in this `Suite`, see also `ids`"""
        return super().problem_names
    @property
    def dimensions(self):
        """list of problem dimensions occuring at least once in this `Suite`"""
        return super().dimensions
    @property
    def number_of_objectives(self):
        """list of number of objectives occuring in this `Suite`"""
        return super().number_of_objectives
    @property
    def indices(self):
        """list of all problem indices, deprecated.
        
        These values are (only) used to call the underlying C structures.
        Indices used in the Python interface run between 0 and `len(self)`.
        """
        return super().indices
    @property
    def name(self):
        """name of this suite as used to instantiate the suite via `Suite(name, ...)`"""
        return super().name
    @property
    def instance(self):
        """instance of this suite as used to instantiate the suite via
        `Suite(name, instance, ...)`"""
        return super().instance
    @property
    def options(self):
        """options for this suite as used to instantiate the suite via
        `Suite(name, instance, options)`"""
        return super().options
    @property
    def info(self):
        return str(self)

class Observer(_Observer):
    """Observer which can be "attached to" one or several problems, however not
    necessarily at the same time.

    The typical observer records data to be used in the COCO post-processing
    module `cocopp` afterwards.

    >>> import cocoex as ex
    >>> suite = ex.Suite("bbob", "", "")
    >>> assert len(suite) == 2160
    >>> f = suite.get_problem(33)
    >>> assert f.id.endswith('f003_i04_d02')
    >>> observer = ex.Observer("bbob",
    ...                        "result_folder: doctest")
    >>> f.observe_with(observer)  # the same as observer.observe(f)  # doctest: +ELLIPSIS
    <cocoex...
    >>> # work work work with observed f
    >>> f.free()

    Details
    -------

        - ``f.free()`` in the above example must be called before to observe
          another problem with the "bbob" observer. Otherwise the Python
          interpreter will crash due to an error raised from the C code.
    
        - Due to technical sublties between Python/Cython/C, the pointer to the
          underlying C observer is passed by global assignment with
          `_update_current_observer_global()`

    """

    def __init__(self, name, options):
        """``options`` can be a string or a `dict`"""
        # this __init__ defines the arguments for _Observer.__cinit__,
        # which is called implicitly
        # super(Observer, self).__init__(name, options)  # fails (only) in Python 3

    def observe(self, problem):
        """`observe(problem)` let `self` observe the `problem: Problem` by
        calling `problem.observe_with(self)`.
        """
        problem.observe_with(self)
        return self

    def signal_restart(self, problem):
        super(Observer, self).signal_restart(problem)

    @property
    def name(self):
        """name of the observer as used with `Observer(name, ...)` to instantiate
        `self` before
        """
        return super().name
    @property
    def options(self):
        return super().options
    @property
    def state(self):
        return super().state
    @property
    def result_folder(self):
        """name of the output folder.

        This name may not be the same as input option `result_folder`.
        """
        return super().result_folder

# this definition is copy-edited from interface, solely to pass docstrings to pydoctor
class Problem(_Problem):
    """`Problem` instances are usually generated using class `Suite`.
    
    The main feature of a problem instance is that it is callable, returning the
    objective function value when called with a candidate solution as input.
    
    It provides other useful properties and methods like `dimension`,
    `number_of_constraints`, `observe_with`, `initial_solution_proposal`...

    """
    def __init__(self):
        super().__init__()
    def constraint(self, x):
        """return constraint values for `x`. 

        By convention, constraints with values <= 0 are satisfied.
        """
        return super().constraint(x)

    def logger_biobj_feed_solution(self, evaluation, y):
        """Feed the given solution to logger_biobj in order to reconstruct its
        output.

        Return 1 if the given solution updated the archive and 0 otherwise.

        Used by preprocessing when updating the .info, .dat and .tdat files
        with new indicator reference values.
        """
        return super().logger_biobj_feed_solution(evaluation, y)


    def add_observer(self, observer):
        """`add_observer(self, observer: Observer)`, see `observe_with`.
        """
        return self.observe_with(observer)

    def observe_with(self, observer):
        """``observe_with(self, observer: Observer)`` attaches an `Observer`
        instance to this problem.

        Attaching an observer can be considered as wrapping the observer
        around the problem. For the observer to be finalized, the problem
        must be free'd (implictly or explicitly).

        Return the observed problem `self`.

        Details: `observer` can be `None`, in which case nothing is done.

        See also: class `Observer`
        """
        return super().observe_with(observer)

    def _f0(self, x):
        """"inofficial" interface to `self` with target f-value of zero. """
        return self(x) - self.final_target_fvalue1

    def initial_solution_proposal(self, restart_number=None):
        """return feasible initial solution proposals.

        For unconstrained problems, the proposal is different for each
        consecutive call without argument and for each `restart_number`
        and may be different under repeated calls with the same
        `restart_number`. ``self.initial_solution_proposal(0)`` is the
        same as ``self.initial_solution``.

        Conceptual example::

            # given: a suite instance, a budget, and fmin
            for problem in suite:
                # restart until budget is (over-)exhausted
                while problem.evaluations < budget and not problem.final_target_hit:
                    fmin(problem, problem.initial_solution_proposal())

        Details: by default, the first proposal is the domain middle or
        the (only) known feasible solution.
        Subsequent proposals are coordinate-wise sampled as the sum
        of two iid random variates uniformly distributed within the
        domain boundaries. On the ``'bbob'`` suite their density is
        0.2 * (x / 5 + 1) for x in [-5, 0] and
        0.2 * (1 - x / 5) for x in [0, 5] and zero otherwise.

        """
        return super().initial_solution_proposal(restart_number)
    @property
    def initial_solution(self):
        """return feasible initial solution"""
        return super().initial_solution()
    @property
    def observers(self):
        """list of observers wrapped around this problem"""
        return super().list_of_observers
    @property
    def is_observed(self):
        """problem ``p`` is observed ``p.is_observed`` times.

        See also: the list of observers in property `observers`.
        """
        return super().is_observed

    @property
    def number_of_variables(self):  # this is cython syntax, not known in Python
        # this is a class definition which is instantiated automatically!?
        """Number of variables this problem instance expects as input."""
        return super().number_of_variables
    @property
    def dimension(self):
        """alias for `number_of_variables` of the input space"""
        return self.number_of_variables
    @property
    def number_of_objectives(self):
        "number of objectives, if equal to 1, call returns a scalar"
        return super().number_of_objectives
    @property
    def number_of_constraints(self):
        "number of constraints"
        return super().number_of_constraints
    @property
    def lower_bounds(self):
        """depending on the test bed, these are not necessarily strict bounds
        """
        return super().lower_bounds
    @property
    def upper_bounds(self):
        """depending on the test bed, these are not necessarily strict bounds
        """
        return super().upper_bounds
    @property
    def evaluations(self):
        """number of times this `Problem` instance was evaluated"""
        return super().evaluations()
    @property
    def final_target_hit(self):
        """return 1 if the final target is known and has been hit, 0 otherwise
        """
        return super().final_target_hit(self.problem)
    @property
    def final_target_fvalue1(self):
        return super().final_target_fvalue1(self.problem)
    @property
    def best_observed_fvalue1(self):
        return super().best_observed_fvalue1()

    def free(self, force=False):
        """Free the given test problem.

        Not strictly necessary (unless, possibly, for the observer). `free`
        ensures that all files associated with the problem are closed as
        soon as possible and any memory is freed. After free()ing the
        problem, all other operations are invalid and will raise an
        exception.
        """
        super().free(force)

    @property
    def id(self):
        "ID as string without spaces or weird characters"
        return super().id

    @property
    def id_function(self):
        "function number inferred from `id`"
        return super().id_function

    @property
    def id_instance(self):
        "instance number inferred from `id`"
        return super().id_instance

    @property
    def id_triple(self):
        "unique problem id triplet ``(function, dimension, instance)``"
        return super().id_triple

    @property
    def name(self):
        """human readible short description with spaces"""
        return super().name

    @property
    def index(self):
        """problem index in the benchmark `Suite` of origin"""
        return super().index

    @property
    def suite(self):
        """benchmark suite this problem is from"""
        return super().suite

    @property
    def info(self):
        """human readible info, alias for ``str(self)``.

        The format of this info string is not guarantied and may change
        in future.

        See also: ``repr(self)``
        """
        return str(self)



import numpy as np

# ===============================================
# the most basic example solver
# ===============================================
def random_search(fun, lbounds, ubounds, budget):
    """Efficient implementation of uniform random search between
    `lbounds` and `ubounds`
    """
    lbounds, ubounds = np.array(lbounds), np.array(ubounds)
    dim, x_min, f_min = len(lbounds), None, None
    max_chunk_size = 1 + 4e4 / dim
    while budget > 0:
        chunk = int(max([1, min([budget, max_chunk_size])]))
        # about five times faster than "for k in range(budget):..."
        X = lbounds + (ubounds - lbounds) * np.random.rand(chunk, dim)
        if fun.number_of_constraints > 0:
            FC = [(fun(x), fun.constraint(x)) for x in X]  # call  objective and constraints
            F = [fc[0] for fc in FC if all(fc[1] <= 0)]
            budget -= chunk  # one more to account for constraint evals
        else:
            F = [fun(x) for x in X]
        if fun.number_of_objectives == 1:
            index = np.argmin(F) if len(F) else None
            if index is not None and (f_min is None or F[index] < f_min):
                x_min, f_min = X[index], F[index]
        budget -= chunk
    return x_min





import pytest
import numpy as np

from cocoex import Suite
from pathlib import Path

TEST_DIR = Path(__file__).parent


def read_test_vectors(fd):
    """
    Read the number of test vectors, followed by the 40D test vectors
    from ${fd}. Return a list of numpy arrays containing the test vectors.
    """
    number_of_test_vectors = int(fd.readline().rstrip())
    ## Preallocate the testvectors list
    test_vectors = number_of_test_vectors * [None]
    for i in range(number_of_test_vectors):
        line = fd.readline().rstrip()
        test_vectors[i] = np.fromstring(line, dtype=float, sep=" ")
    return test_vectors


def test_bbob2009():
    with open(TEST_DIR / "bbob2009_testcases.txt") as fd:
        test_suite = fd.readline().rstrip()
        test_vectors = read_test_vectors(fd)
      
        suite = Suite(test_suite, "instances:1-15", "")
        previous_problem_index = -1
        for test_case in fd:
            ## A test case is a 4-tuple (deprecated_problem_index, problem_index, test_vector_id,
            ## expected_y) separated by a tab.
            deprecated_problem_index, problem_index, test_vector_id, expected_y = test_case.split()

            ## We cache the problem instances because creating an instance
            ## can be expensive depending on the transformation.
            if problem_index != previous_problem_index:
                problem = suite.get_problem(int(problem_index))
                previous_problem_index = problem_index
            test_vector = test_vectors[int(test_vector_id)]
            y = problem(test_vector[:problem.number_of_variables])
            assert y == pytest.approx(float(expected_y))



# -*- mode: cython -*-
#cython: language_level=3, boundscheck=False, c_string_type=str, c_string_encoding=ascii

import numpy as np
cimport numpy as np

from .exceptions import NoSuchProblemException, NoSuchSuiteException

np.import_array()


cdef extern from 'coco.h':
    ctypedef struct coco_problem_t:
        pass

    char *coco_problem_get_id(coco_problem_t *p)
    void coco_evaluate_function(coco_problem_t *p, double *x, double *y)
    void coco_problem_free(coco_problem_t *p)
    void bbob_problem_best_parameter_print(coco_problem_t *p)
    
    int coco_problem_get_best_parameter(const coco_problem_t *problem, double *best_parameter)
    int coco_problem_is_tainted(const coco_problem_t *problem)

    int coco_problem_get_best_parameter(const coco_problem_t *problem, double *best_parameter)

# IMPORTANT: These functions are *not* declared public in coco.h so we have to
# explicitly declare it as an external function. Otherwise Cython will *not*
# add a declaration to the generated source files.
cdef extern coco_problem_t *coco_get_bbob_problem(size_t function, size_t
                                                  dimension, size_t instance)
cdef extern double coco_problem_get_best_value(coco_problem_t *p)


cdef class BareProblem:
    """A bare benchmark function from one of the available suites.

    Examples
    --------

    >>> import numpy as np

    Create a 13 dimensional sphere function

    >>> fn = BareProblem("bbob", 1, 13, 1)
    >>> fn
    BareProblem('bbob', 1, 13, 1)

    We can also get a short mnemonic name for the function

    >>> str(fn)
    'bbob_f001_i01_d13'

    And of course evaluate it

    >>> x = np.zeros(fn.dimension)
    >>> fn(x)
    124.61122368000001
    """

    cdef coco_problem_t *_problem
    cdef readonly char* suite_name
    cdef readonly char* id
    cdef readonly int function
    cdef readonly int dimension
    cdef readonly int instance

    def __init__(self, suite_name: str, function: int, dimension: int, instance: int):
        """
        Create a bare benchmark problem from one of the COCO suites.

        Parameters
        ----------
        suite_name
            Name of benchmark suite ("bbob" only currently)

        function
            ID of function from suite

        dimension
            Dimension of instantiated function

        instance
            Instance ID of instantiated function


        Raises
        ------
        NoSuchSuiteException
           If the `suite_name` is not known or not yet supported

        NoSuchProblemException
          If no problem with the given `function`, `dimension` and `instance` exists in
          the given `suite_name`.
        """
        self.suite_name = suite_name
        self.function = function
        self.dimension = dimension
        self.instance = instance
        self._problem = NULL
        if suite_name == "bbob":
            self._problem = coco_get_bbob_problem(function, dimension, instance)
            if self._problem == NULL:
                # FIXME: Possibly extend Exception to include dimension and instance?
                raise NoSuchProblemException(suite_name, function)
        else:
            raise NoSuchSuiteException(suite_name)

        self.id = coco_problem_get_id(self._problem)

    def __del__(self):
        if self._problem != NULL:
            coco_problem_free(self._problem)

    def best_value(self):
        """Return the best (lowest) possible function value"""
        return coco_problem_get_best_value(self._problem)

    def best_parameter(self):            
        cdef double[::1] x
        x = np.zeros(self.dimension, dtype=np.float64)

        res = coco_problem_get_best_parameter(self._problem, &x[0])
        if res == 0: # Failure
            return None # FIXME: Raise an exception?
        else:
            return np.asarray(x)

    @property
    def is_tainted(self):
        is_tainted = coco_problem_is_tainted(self._problem)
        return is_tainted != 0
    
    def __str__(self):
        return self.id

    def __repr__(self):
        return f"BareProblem('{self.suite_name}', {self.function}, {self.dimension}, {self.instance})"

    def __call__(self, x):
        cdef double[::1] xi
        cdef double[:, ::1] X
        cdef Py_ssize_t N, D, i
        cdef double[::1] y_view
        cdef double y
        if isinstance(x, list):
            x = np.array(x, dtype=float)
        if x.ndim == 1:
            # Evaluate a single parameter
            xi = np.array(x, dtype=float)
            coco_evaluate_function(self._problem, &xi[0], &y)
            return y
        elif x.ndim == 2:
            # Evaluate several parameters at once
            X = x
            N = X.shape[0]
            D = X.shape[1]
            Y = np.zeros(N, dtype=np.float64)
            y_view = Y
            for i in range(N):
                xi = X[i, :]
                coco_evaluate_function(self._problem, &xi[0], &y)
                y_view[i] = y
            return Y
        else:
          return None




import pytest
import cocoex
from cocoex.exceptions import NoSuchSuiteException


def test_crash_no_dimension():
    """C code crashes when no dimensions remain after filtering.

    See #2181"""
    with pytest.raises(NoSuchSuiteException):
        cocoex.Suite("bbob", "", "dimensions:4")




class NoSuchProblemException(Exception):
    def __init__(self, suite, function_id):
        self.suite = suite
        self.function_id = function_id

    def __str__(self):
        s = "Problem suite '{suite}' lacks a function with id '{id}'."
        return s.format(suite=self.suite, id=self.function_id)

    def __repr__(self):
        s = "NoSuchProblemException('{suite}', {id})"
        return s.format(suite=self.suite, id=self.function_id)


class NoSuchSuiteException(Exception):
    def __init__(self, suite):
        self.suite = suite

    def __str__(self):
        return f"Unknown benchmark suite '{self.suite}'."

    def __repr__(self):
        return f"NoSuchSuiteException('{self.suite}')"


class InvalidProblemException(Exception):
    def __init__(self):
        pass



import ast as _ast  # avoid name space polluting
import os as _os
import sys as _sys
import time as _time
import warnings as _warnings
import collections as _collections  # defaultdict, OrderedDict

import numpy as np


def about_equal(a, b, precision=1e-6):
    """
    Return True if the floating point number ${a} and ${b} are about equal.
    """
    if a == b:
        return True

    absolute_error = abs(a - b)
    larger = a if abs(a) > abs(b) else b
    relative_error = abs(a - b) / (abs(larger) + 2 * _sys.float_info.min)

    if absolute_error < (2 * _sys.float_info.min):
        return True
    return relative_error < precision


def args_to_dict(args, known_names, specials=None, split='=',
                 print=lambda *args, **kwargs: None):
    """return a `dict` from a list of ``"name=value"`` strings.

    `args` come in the form of a `list` of ``"name=value"`` strings
    without spaces, like ``["budget_multiplier=100"]``.

    Return ``dict(arg.split(split) for arg in args)`` in the most
    basic case, but additionally (i) checks that the keys of this
    `dict` are known names, (ii) evaluates the values in some cases
    and (iii) handles `specials`.

    `known_names` is an iterable (`dict` or `list` or `tuple`) of strings.
    If ``known_names is None``, all args are processed, otherwise a
    `ValueError` is raised for unknown names. This is useful if we
    want to re-assign variables (overwrite default values) and avoid
    spelling mistakes pass silently.

    The value is processed as a Python literal with `ast.literal_eval`
    or remains a `str` when this is unsuccessful.

    `specials` is a `dict` and can currently only contain ``'batch'``,
    followed by ``"name1/name2"`` as value. ``name1`` and ``name2`` are
    then assigned from the values in `arg`, for example to 2 and 4 with
    ``batch=2/4``.

    A main usecase is to process ``sys.argv[1:]`` into a `dict` in a
    python script, like::

        command_line_dict = args_to_dict(sys.argv[1:], globals())
        globals().update(command_line_dict)

    >>> import cocoex
    >>> d = cocoex.utilities.args_to_dict(["budget=2.3", "bed=bed-name", "number=4"],
    ...                                   ["budget", "bed", "number", "whatever"])
    >>> len(d)
    3
    >>> assert d['bed'] == 'bed-name'
    >>> assert isinstance(d["budget"], float)

    """
    def eval_value(value):
        try:
            return _ast.literal_eval(value)
        except Exception:  # ValueError or SyntaxError or ??
            return value
    res = {}
    for arg in args:
        name, value = arg.split(split)
        # what remains to be done is to verify name,
        # compute non-string value, and assign res[name] = value
        if specials and name in specials:
            if name == 'batch':
                print('batch:')
                if len(specials['batch'].split('/')) != len(value.split('/')):
                    raise ValueError("'{0}' is not a valid value for argument 'batch={0}'."
                                     " A valid example is 'batch=1/3' ({1})"
                                     .format(value, specials['batch']))  # zip(..., strict=...) bails in Python3.9
                for k, v in zip(specials['batch'].split('/'), value.split('/')):
                    res[k] = int(v)  # batch accepts only int
                    print(' ', k, '=', res[k])
                continue  # name is processed
            else:
                raise ValueError(name, 'is unknown special')
        for known_name in known_names if known_names is not None else [name]:
            # check that name is an abbreviation of known_name and unique in known_names
            if known_name.startswith(name) and (
                        sum([other.startswith(name)
                             for other in known_names or [name]]) == 1):
                if known_name in res:
                    if res[known_name] != eval_value(value):
                        raise ValueError("found two values for argument {0}={1} and ={2}"
                                         .format(known_name, res[known_name], value))
                else:
                    res[known_name] = eval_value(value)
                print(known_name, '=', res[known_name])
                break  # name == arg.split()[0] is processed
        else:
            raise ValueError('Argument name "{}" is ambiguous or not given in ``known_names=={}``'
                             .format(name, sorted([k for k in known_names if not k.startswith('_')
                                    and k not in ('division', 'print_function', 'unicode_literals',
                                                  'sys', 'warnings', 'time', 'defaultdict', 'os', 'np', 'scipy')])))
    return res

def dict_to_eval(dict_, ignore_list=('_', 'self')):
    """return a pruned `dict` so that ``ast.literal_eval(repr(dict_))`` works.

    Keys that start with entries from `ignore_list` when interpreted as
    `str` are removed, by default those starting with ``_`` or ``self``,
    where `ignore_list` must be a `str` or a `tuple` of `str`.

    See also `write_setting`.
    """
    res = {}
    for k in dict_:
        if str(k).startswith(ignore_list):
            continue
        try:
            _ast.literal_eval(repr(dict_[k]))
        except Exception:
            pass
        else:
            res[k] = dict_[k]
    return res

def write_setting(dict_, filename, ignore_list=None):
    """write a simplified parameters dictionary to a file,

    for keeping a record or, e.g., for checking the `budget_multiplier`
    later.

    A typical usecase is ``write_setting(locals(), 'parameters.pydat')``.

    When ``ignore_list is not None`` it is passed to `dict_to_eval`
    which determines which parameters are written or omitted. By default,
    keys starting with ``_`` or ``self`` are omitted and items that bail on
    `literal_eval`.

    See also `dict_to_eval` and `read_setting`.
    """
    if isinstance(filename, (tuple, list)):
        filename = _os.path.join(*filename)
    if not _os.path.exists(filename):
        with open(filename, 'wt') as f:
            if ignore_list is None:
                f.write(repr(dict_to_eval(dict_)))
            else:
                f.write(repr(dict_to_eval(dict_, ignore_list)))
    else:
        _warnings.warn('nothing written as the file "' + filename +
                       '"\nexists already (this should never happen as\n'
                       'each experiment should be written in a new (sub-)folder)')

def read_setting(filename, warn=True):
    """return file content evaluated as Python literal (e.g. a `dict`),

    return `None` if `filename` is not a valid path.

    If `warn`, throw a warning when the file `filename` does not exist.

    A typical usecase could be
    ``old_multiplier = read_setting('parameters.pydat')['budget_multiplier']``.

    See also `write_setting`.
    """
    if _os.path.exists(filename):
        with open(filename, 'rt') as f:
            return _ast.literal_eval(f.read())
    warn and _warnings.warn("Parameter file '{}' (to check setting consistency)"
                            " does not exist".format(filename))

def forgiving_import(module, warn_level=0):
    """Do nothing if import fails, return the imported module otherwise.

    Usage::

        cma = forgiving_import('cma')

    in place of::

        import cma

    This is helpful to keep some code smoothly working when `cma` is not
    installed and not used in the current use case.
    """
    try:
        import importlib
    except ImportError:
        print('Please replace \n\n   {}\n\n with \n\n    {}\n\n'
              'or incomment the respective line.'.format(
            "module = forgiving_import('module')", 
            "import module"))
        raise
    try:
        return importlib.import_module(module)
    except ImportError:
        if warn_level:
            print("importing module `{}` with `importlib` failed".format(module))
        if warn_level > 1:
            raise

class ObserverOptions(dict):
    """a `dict` with observer options which can be passed to
    the (C-based) `Observer` via the `as_string` property.

    See http://numbbo.github.io/coco-doc/C/#observer-parameters
    for details on the available (C-based) options.

    Details: When the `Observer` class in future accepts a dictionary
    also, this class becomes superfluous and could be replaced by a method
    `default_observer_options` similar to `default_observers`.
    """
    def __init__(self, options={}):
        """set default options from global variables and input ``options``.

        Default values are created "dynamically" based on the setting
        of module-wide variables `SOLVER`, `suite_name`, and `budget`.
        """
        dict.__init__(self, options)
    def update(self, *args, **kwargs):
        """add or update options"""
        dict.update(self, *args, **kwargs)
        return self
    def update_gracefully(self, options):
        """update from each entry of parameter ``options: dict`` but only
        if key is not already present
        """
        for key in options:
            if key not in self:
                self[key] = options[key]
        return self
    @property
    def as_string(self):
        """string representation which is accepted by `Observer` class,
        which calls the underlying C interface
        """
        s = str(self).replace(',', ' ')
        for c in ["u'", "'", "{", "}"]:
            s = s.replace(c, '')
        return s


class ProblemNonAnytime:
    """The non-anytime problem class.

    Serves to benchmark a "budgeted" algorithm whose behavior decisively
    depends on a budget input parameter.

    Usage::

        # given: suite and observer instances, budget_list, fmin
        for index in range(len(suite)):
            with ProblemNonAnytime(suite, observer, index) as problem:
                x0 = problem.initial_solution  # or whatever fmin needs as input
                for budget in sorted(budget_list):
                    x = fmin(problem, x0, budget)  # minimize
                    problem.delivered(x)  # prepares for next budget also
                    if problem.final_target_hit:
                        break

    Details: This class maintains two problems - one observed and the
    other unobserved. It switches from unobserved to observed when
    ``p_unobserved.evaluations >= p_observed.evaluations``.

    """
    inherited_constant_attributes = [
            'dimension',
            'id',
            'index',
            'info',
            'initial_solution',
            'lower_bounds',
            'name',
            'number_of_constraints',
            'number_of_objectives',
            'number_of_variables',
            'upper_bounds']

    def __init__(self, suite, observer, index):
        self.suite = suite
        self.p_unobserved = suite[index]
        self.p_observed = suite[index].add_observer(observer)
        self._p = self.p_observed
        self.evaluations = 0
        self._number_of_delivery_evaluations = 0  # just FTR
        for key in ProblemNonAnytime.inherited_constant_attributes:
            setattr(self, key, getattr(self._p, key))

    def __call__(self, x, *args, **kwargs):
        if self.evaluations > self.p_observed.evaluations:
            raise ValueError(
                "Evaluations {0} are larger than observed evaluations {1}"
                "".format(self.evaluations, self.p_observed.evaluations))
        if self.evaluations >= self.p_observed.evaluations:
            self._p = self.p_observed
        self.evaluations += 1
        return self._p(x, *args, **kwargs)

    def delivered(self, x, *args, **kwargs):
        """to be called with the solution returned by the solver.

        The method records the delivered solution (if necessary) and
        prepares the problem to be run on the next budget by calling
        `reset`.
        """
        if self.evaluations < self.p_observed.evaluations:
            raise ValueError(
                "Delivered solutions should come from increasing budgets,\n"
                "however the current budget = {0} is smaller than the "
                "delivery budget = {1}".format(
                    self.p_observed.evaluations, self.evaluations))
        # assuming that "same fitness" means the solution was already evaluated
        if self.p_unobserved(x, *args, **kwargs) != self.best_observed_fvalue1:
            # ideally we would decrease the observed evaluation counter,
            # but instead we now increase it only one time step later in
            # __call__. That is, in effect, we exchange the next solution
            # to be observed with this delivered solution `x`.
            self.p_observed(x, *args, **kwargs)
            self._number_of_delivery_evaluations += 1  # just FTR
        self.reset()

    def reset(self):
        """prepare the problem to be run on the next budget"""
        self.evaluations = 0
        self._p = self.p_unobserved

    def free(self):
        self.p_observed.free()
        self.p_unobserved.free()

    def print_progress(self):
        if not self.index % (len(self.suite) / len(self.suite.dimensions)):
            print(f"\nd={self.dimension}: ")
        print("{}{} ".format(
            self.id[self.id.index("_f") + 1:self.id.index("_i")],
            self.id[self.id.index("_i"):self.id.index("_d")]), end="",
              flush=True)
        if not (self.index + 1) % 10:
            print("")

    def __enter__(self):
        return self

    def __exit__(self, *args):
        self.free()

    @property
    def final_target_hit(self):
        return self.p_observed.final_target_hit

    @property
    def best_observed_fvalue1(self):
        return self.p_observed.best_observed_fvalue1

    @property
    def random_solution(self):
        return self.lower_bounds + (
            (np.random.rand(self.dimension) + np.random.rand(self.dimension))
            * (self.upper_bounds - self.lower_bounds) / 2)


class SameFunction:
    """Count the number of consecutive instances of the same function.

    Useful to limit the number of repetitions on the same function.

    Example:

    >>> import cocoex
    >>> from cocoex.utilities import SameFunction
    >>>
    >>> suite = cocoex.Suite('bbob', '', '')
    >>> already_seen = SameFunction()
    >>> processed = 0
    >>> for problem in suite:
    ...     if already_seen(problem.id) > 5:
    ...         continue
    ...     # do something here only with the first five instances
    ...     processed += 1
    >>> processed, len(suite), already_seen.count
    (864, 2160, 15)

    More arbitrary tests:

    >>> for i in range(4):
    ...     if seen('f001_i%d' % i) > 2:
    ...         continue
    ...     # do something here only the first two instances
    >>> seen.count
    4
    >>> seen('f_d03_i001')
    0
    >>> seen('f_d03_i02')
    1
    >>> for i in range(4):
    ...     if seen('f%d_i%d' % (i, i)):
    ...         break
    >>> i, seen.count
    (3, 1)

    """
    @staticmethod
    def filter(id):
        """remove instance information and return a `tuple`"""
        return tuple(i for i in id.split('_') if not i.startswith('i'))
    def __init__(self):
        self.count = 0
    def __call__(self, id):
        """return number of directly preceding calls with similar `id`
        """
        new = SameFunction.filter(id)
        if self.count == 0 or new != self.last:
            self.last = new
            self.count = 0
        self.count += 1
        return self.count - 1  # return old count


class MiniPrint:
    """print dimension when changed and a single symbol for each call.

    Details: print '|' if ``problem.final_target_hit``, ':' if restarted
    and '.' otherwise.
    """
    def __init__(self):
        self.dimension = None
        self.id_function = None
        self.last_index = -1
        self._calls = 0  # formatting aid
        self._functions = 0  # not in use
        self._sweeps = 0
        self._day0 = _time.localtime()[2]
    @property
    def stime(self):
        """current time as string +days since started"""
        ltime = _time.localtime()[2:6]
        s = "%dh%02d:%02d" % tuple(ltime[1:])
        if ltime[0] > self._day0:
            s = s + "+%dd" % (ltime[0] - self._day0)
        return s
    def __call__(self, problem, final=False, restarted=False, final_message=False):
        new_dimension = self.dimension != problem.dimension or (
            self.last_index > problem.index)
        if new_dimension:
            _sys.stdout.flush()
            if self.dimension is not None:
                print('')
            print("%dD %s" % (problem.dimension, self.stime))
            self.dimension = problem.dimension
            self._calls = 0
            self._functions = 0
        # elif not self._calls % 10:
        #     if self._calls % 50:
        #         print(' ', end='')
        #     else:
        #         print()
        if self.id_function != problem.id_function or new_dimension:
            self.id_function = problem.id_function
            if not new_dimension and self._calls >= 50:
                self._calls = 0
                print()
            print('{}f{}'.format(' ' if self.id_function < 10 else '',
                                 self.id_function), end='', flush=True)
            self._functions += 1
        self.last_index = problem.index
        self._calls += 1
        print('|' if problem.final_target_hit else ':' if restarted else '.', end='')
        if final:  # final print
            self._sweeps += 1
            if final_message:
                try:
                    len(final_message)  # poor mans string type check
                    print(final_message)
                except TypeError:
                    print('\nSuite done ({})'.format(self._sweeps))
        _sys.stdout.flush()


class ShortInfo:
    """print minimal info during benchmarking.

    After initialization, to be called right before the solver is called with
    the respective problem. Prints nothing if only the instance id changed.

    Example output:

        Jan20 18h27:56, d=2, running: f01f02f03f04f05f06f07f08f09f10f11f12f13f14f15f16f17f18f19f20f21f22f23f24f25f26f27f28f29f30f31f32f33f34f35f36f37f38f39f40f41f42f43f44f45f46f47f48f49f50f51f52f53f54f55 done

        Jan20 18h27:56, d=3, running: f01f02f03f04f05f06f07f08f09f10f11f12f13f14f15f16f17f18f19f20f21f22f23f24f25f26f27f28f29f30f31f32f33f34f35f36f37f38f39f40f41f42f43f44f45f46f47f48f49f50f51f52f53f54f55 done

        Jan20 18h27:57, d=5, running: f01f02f03f04f05f06f07f08f09f10f11f12f13f14f15f16f17f18f19f20f21f22f23f24f25f26f27f28f29f30f31f32f33f34f35f36f37f38f39f40f41f42f43f44f45f46f47f48f49f50f51f52f53f54f55 done

    """
    def __init__(self):
        self.f_current = None  # function id (not problem id)
        self.d_current = 0  # dimension
        self.t0_dimension = _time.time()
        self.evals_dimension = 0
        self.evals_by_dimension = {}
        self.runs_function = 0

    def print(self, problem, end="", **kwargs):
        print(self(problem), end=end, **kwargs)
        _sys.stdout.flush()

    def add_evals(self, evals, runs):
        self.evals_dimension += evals
        self.runs_function += runs

    def dimension_done(self):
        self.evals_by_dimension[self.d_current] = (_time.time() - self.t0_dimension) / self.evals_dimension
        s = '\n    %d-D done in %.1e seconds/evaluation' % (self.d_current, self.evals_by_dimension[self.d_current])
        # print(self.evals_dimension)
        self.evals_dimension = 0
        self.t0_dimension = _time.time()
        return s

    def function_done(self):
        s = "(%d)" % self.runs_function + (2 - int(np.log10(self.runs_function))) * ' '
        self.runs_function = 0
        return s

    def __call__(self, problem):
        """uses `problem.id` and `problem.dimension` to decide what to print.
        """
        f = "f" + problem.id.lower().split('_f')[1].split('_')[0]
        res = ""
        ran_once = self.f_current is not None
        f_changed = f != self.f_current
        d_changed = problem.dimension != self.d_current
        run_complete = f_changed or d_changed
        if ran_once and run_complete:
            res += self.function_done() + ' '
        if d_changed:
            res += '%s%s, d=%d, running: ' % (
                        self.dimension_done() + "\n\n" if self.d_current else '',
                        ShortInfo.short_time_stap(),
                        problem.dimension)
            self.d_current = problem.dimension
        if run_complete:
            res += '%s' % f
        if f_changed:
            self.f_current = f
        # print_flush(res)
        return res

    def print_timings(self):
        print("  dimension seconds/evaluations")
        print("  -----------------------------")
        for dim in sorted(self.evals_by_dimension):
            print("    %3d      %.1e " %
                  (dim, self.evals_by_dimension[dim]))
        print("  -----------------------------")

    @staticmethod
    def short_time_stap():
        t = _time.asctime().split()
        d = t[0]
        d = t[1] + t[2]
        h, m, s = t[3].split(':')
        return d + ' ' + h + 'h' + m + ':' + s


def ascetime(sec, decimals=0):
    """return elapsed time as str.

    Example: return ``"0h33:21" if sec == 33*60 + 21``.
    """
    if sec < 0:  # sort-of ValueError?
        return "%.1f sec" % (sec)
    ds = str(sec % 1)[1:2+decimals] if decimals else ""  # fraction
    # if sec < 10:  # breaks backwards compatibility of processing printed output
    #     return "%d%s sec" % (sec, str(sec % 1)[1:2+max((1, decimals))])
    if sec < 100 and decimals:
        return "%d%s sec" % (sec, ds)
    h = sec / 60**2  # hours
    m = 60 * (h % 1)  # minutes
    s = 60 * (m % 1)  # seconds
    return "%dh%02d:%02d%s" % (h, m, s, ds)


def print_flush(*args):
    """print without newline but with flush"""
    print(*args, end="")
    _sys.stdout.flush()

class ExperimentRepeater:
    """Allow to automatically repeat an experiment based on budget and successes.

    The class tracks evaluations and the `final_target_hit` status
    ("successes") of the problem instances of a `cocoex.Suite`. Based on
    the stored information, the `done` method allows to run sub-experiment
    sweeps until some given budget is exhaused, namely, ``budget_multiplier
    * dimension``.

    Example code snippet::

        [...]
        repeater = cocoex.ExperimentRepeater(budget_multiplier)
        [...]
        while not repeater.done():
            for problem in suite:
                if repeater.done(problem):
                    continue
                problem.observe_with(observer)  # generates the data for cocopp postprocessing
                fmin(problem, repeater.initial_solution_proposal(problem))
                repeater.track(problem)  # record evaluations and success

    The call ``repeater.done()`` checks whether the budget is exhausted or
    enough trials were successful on all problems, or an instance lacks a
    trial on some problems, it is akin to, but not exactly the same as,
    ``all(repeater.done(p) for p in suite)``. ``ExperimentRepeater(0)``
    does exactly one sweep and no repetitions.

    The stored data can be queried, in which case `problem` can be a
    ``(id_function, dimension, id_instance)`` tuple. In particular, the
    methods `evaluations`, `successes` and `trials` return a `list` of the
    respective data for each instance of `instances`. The last entry of the
    argument tuple is ignored in these calls and can be `None` or omitted.
    The `all` method allows to query all `(id_function, dimension)` entries
    at once::

        >> evals = repeater.all(repeater.evaluations)  # returns a `dict`
        >> fun, dim = 1, 10
        >> evals[(fun, dim)] == repeater.evaluations((fun, dim, None))
        True

    Terminology: a "problem instance" is defined by the triple
    ``(id_function, dimension, id_instance)`` and "problem" may refer
    to any or all problem instances with the same ``(id_function,
    dimension)``.

    Details
    -------
    Only the `track` method relies on passing a `cocoex.Problem`
    instance, otherwise a ``(id_function, dimension, id_instance)``
    tuple is eligible. The tracked information is stored in the `._data`
    dictionary with ``(id_function, dimension)`` as keys. The `data` method
    implements safe access to this dictionary.

    When problem instances are repeated in a single suite, they may be
    *partially* skipped *after* the first full sweep. That is, the
    configuration ``1-5,1-5,1-5`` can also lead to four trials of each
    instance 1-5, because all instances have been repeated the same number
    of times.

    >>> import cocoex  # some not very meaningful testing
    >>> repeater = cocoex.ExperimentRepeater(2)
    >>> assert not repeater.done((1, 2))
    >>> assert repeater._sweeps == 0 and not repeater.done() and repeater._sweeps == 1
    >>> assert len(repeater.data((1, 10))) == 0
    >>> assert len(repeater.evaluations((1, 2))) == 0

    """
    def __init__(self, budget_multiplier, min_successes=11):
        """``min_successes=11`` for instances 1-5 provokes at least three

        sweeps, hence 15 trials, given the `budget_multiplier` is large
        enough and the algorithm terminates early enough before the budget
        is exhausted. 3 x 1-5 is the instance-setup from BBOB 2009, however
        the choice of the initial solution still slightly differs.
        """
        self.params = {k: v for k, v in locals().items() if k != 'self'}
        self.max_sweeps = 1e4  # should not be necessary
        self._dimension_offset = 0
        '''budget = (dimension + offset) * budget_multiplier'''
        self._data = _collections.OrderedDict()
        '''a ``dict[(fun, dim)]`` of ``dict[iinst]`` of `list` of
           ``(evaluations, success)`` `tuples`'''
        self._calls = 0  # calls of `track`
        self._sweeps = 0  # calls of ``done()`` without argument
    def all(self, method):
        """return `dict` with ``method(key)`` as values and

        with all ``(id_function, dimension)`` as keys.

        Works with the instance methods `evaluations`, `successes`,
        `budget_exhausted`, `trials`, or `instances` as argument or the
        respective strings.
        """
        if method == str(method):
            method = getattr(self, method)
        return {(fun, dim): method((fun, dim)) for (fun, dim) in self._data}
    def budget_from_dimension(self, problem_or_dimension):
        """return budget_multiplier x ((problem.dimension or dimension) + self.offset).

        By default, ``offset == 0``.
        """
        
        dimension = problem_or_dimension if isinstance(problem_or_dimension, int) else (
                    self.problem_to_ids(problem_or_dimension)[1])
        return (dimension + self._dimension_offset) * self.params['budget_multiplier']
    def _empty_data(self, problem=None):
        """should always return an empty list, called in the `data` method"""
        if problem is None:
            return {p: self._empty_data(p) for p in self._data if self._empty_data(p)}
        fun, dim = self.problem_to_ids(problem)[:2]
        if (fun, dim) in self._data:
            d = self._data[(fun, dim)]
            return [ii for ii in d if not d[ii]]  # list of instance numbers with no data
        return []
    @property
    def n_problems(self):
        """current number of tracked problems (different ``(id_function, dimension)``)"""
        return sum(len(self.trials(k)) > 0 for k in self._data)
    def _n_instance_data(self, problem):
        return len(self.data_of_instance(problem))
    def n_problem_instances(self, problem=None):
        """number of problem instances (of `problem`) with at least one trial
        """
        if problem is not None:
            return sum(n > 0 for n in self.trials(problem))
            return len(self.data(problem))
            return sum(len(res_list) > 0 for res_list in self.data(problem).values())
        return sum(self.n_problem_instances(p) for p in self._data)
        return sum(len(d) for d in self._data.values())
    def n_trials(self, problem=None):
        """current number of tracked trials from all instances (of `problem`)"""
        if problem is not None:
            return sum(self.trials(problem))
        return sum(self.all(self.n_trials).values())
    def track(self, problem):
        """record problem instance evaluations and success (target hit)"""
        fun, dim, iinst = self.problem_to_ids(problem)
        if (fun, dim) not in self._data:
            self._data[(fun, dim)] = _collections.defaultdict(list)
        self._data[(fun, dim)][iinst] += [(problem.evaluations, problem.final_target_hit)]
        self._calls += 1
    def problem_to_ids(self, problem):
        """return ``(id_function, dimension, id_instance)`` of `problem`"""
        self._check(problem)
        if (not callable(problem)
                and isinstance(problem, (tuple, list))
                and len(problem) in (2, 3)):
            return tuple(problem) if len(problem) == 3 else (problem[0], problem[1], None)
        return problem.id_function, problem.dimension, problem.id_instance
    def _to_key(self, problem):
        """return `data`-key of problem, namely ``(id_function, dimension)``"""
        return self.problem_to_ids(problem)[:2]
    def _check(self, problem):
        """check problem to avoid kernel crashes, raise ValueError"""
        if hasattr(problem, 'info') and 'invalid' in problem.info:
            raise ValueError(problem.info)
    def _assert_consistencies(self):
        if self.done():  # the below is necessary but not sufficient to be done
            assert all(sum(v) for v in zip(self.all('budget_exhausted').values(),
                                           self.all('succeeded').values()))
        for (problem, data) in self._data.items():
            assert self.trials(problem) == [len(data[ii]) for ii in sorted(data)]
            assert sum(n > 0 for n in self.trials(problem)) == len(self.data(problem)), (
                    "found unexpected empty data in problem {}:"
                    "\n  trials=={}\n  data=={}"
                    .format(problem, self.trials(problem), self.data(problem)))
        assert (sum(len(d) for d in self._data.values()) == 
                sum(self.n_problem_instances(p) for p in self._data))
    def data(self, problem):
        """return reference to `dict` with instance as keys,

        where nonexisting entries are returned as empty but not permanently
        created.
        """
        if self._empty_data():
            _warnings.warn("empty data found when calling `data` method"
                           " with problem {}:\n  {}"
                           .format(self.problem_to_ids(problem), self._empty_data()))
        return self._data.get(self._to_key(problem), _collections.defaultdict(tuple))
        # return self._data.[(fun, dim)] if (fun, dim) in self._data else _collections.defaultdict(tuple)  # is minimally quicker
    def data_of_instance(self, problem):
        """return `list` of ``(evaluations, success)`` of problem instance"""
        return self.data(problem).get(self.problem_to_ids(problem)[2], ())
    def instances(self, problem):
        """return `list` of instance IDs tracked for (id_function, dimension) of `problem`.

        Instance IDs are the keys for the ``self.data(problem)`` dictionary.
        """
        return sorted(self.data(problem))  # unsorted dict keys have arbitrary order
    def trials(self, problem, instance=None):
        """return `list` of number of recorded trials for each instance

        if ``instance is None``, otherwise for the given `instance` only.
        """
        dd = self.data(problem)  # dictionary of all data related to problem
        if instance is not None:
            return len(dd[instance])
        return [len(dd[iinst]) for iinst in sorted(dd)]
    def successes(self, problem):
        """return `list` of successes per instance"""
        dd = self.data(problem)  # dictionary of all data related to problem
        return [sum(i[1] for i in dd[key]) for key in sorted(dd)]
    def evaluations(self, problem):
        """return `list` of summed evaluations per problem instance"""
        dd = self.data(problem)  # dictionary of all data related to problem
        return [sum(i[0] for i in dd[key]) for key in sorted(dd)]
    def budget_exhausted(self, problem):
        """compare average evaluations per instance with `budget_from_dimension`"""
        return sum(self.evaluations(problem)) >= max((1, len(self.instances(problem)))
                    ) * self.budget_from_dimension(problem)
    def succeeded(self, problem):
        """return `True` iff problem had at least `min_success` successes"""
        return sum(self.successes(problem)) >= self.params['min_successes']
    def message_sweep(self):
        """return status message and '' when nothing was tracked yet"""
        if not len(self._data):
            if self._sweeps <= 1:
                return ''
            return '\nSweep {} done but no data were recorded'.format(self._sweeps)
        successes = [sum(v) for v in self.all(self.successes).values()]
        succeeded = self.all(self.succeeded).values()
        return ('\nSweep {} done ({} success{} on {} problem{} where {} had >= 1'
            ' success and {} had >= {} successes and {} exhausted the budget)\n'
            .format(self._sweeps,
                    sum(successes),
                    'es' if sum(successes) != 1 else '',
                    len(self._data),
                    's' if len(self._data) != 1 else '',
                    sum(s > 0 for s in successes),
                    sum(succeeded),
                    self.params['min_successes'],
                    sum(b and not s for (s, b) in
                        zip(succeeded, self.all(self.budget_exhausted).values()))))
    def done(self, problem=None, message=True):
        """return `False` iff this/any `problem` instance remains to be (re-)run.

        When ``problem is not None``, return `True` if this problem
        instance does not require another trial, however `False` as long as
        `done` has not been called twice without argument (thereby assuming
        that the loop starts with ``while repeater.done()``).
 
        When ``problem is None``, return `True` if no single recorded problem
        requires another trial.

        Return invariably `False` before the second call of ``done()``
        without argument, which is considered to happen right before the
        second sweep when the first sweep is finished.

        Details
        -------
        ``done()`` without argument gives only consistent results before or
        after a *full* first sweep. In particular, *during* the first sweep
        it cannot account for problems that have not yet been run once.

        Calling ``done()`` increments the sweep counter iff it returns
        `False`, the default for ``max_sweeps`` is ``1e4``. The attribute
        can be directly reassigned at any time.

        See also: `remaining_problems`
        """
        if problem is not None:
            if self._sweeps <= 1 or self.missing_trials(problem) or (
                    not self.succeeded(problem)
                    and not self.budget_exhausted(problem)):
                return False
            return True
        if message:  # prints nothing when no data are found
            print(self.message_sweep(), end='')  # indicates "sequence of problems" case
        self._sweeps += 1  # the sweep number to come
        done = self._sweeps >= self.max_sweeps or (
                self._sweeps > 1 and len(self._data) == 0) or (
                self.n_problem_instances() > 0 and not self.remaining_problems())
        self._sweeps -= done  # don't increment when done
        return done
    def remaining_problems(self, suite=None):
        """return `list` of (probably) remaining problem instances to run.

        `suite`, when given, compares the recorded experiments with the
        problem instances in the suite to determine what remains. The
        `suite` argument is bound to give unexpected results when some of
        the problems from `suite` are skipped to be run with other batches,
        e.g., in parallel.

        See also: `done`
        """
        if suite is not None:
            res = [self.problem_to_ids(p) for p in suite if not self.done(p)]
        else:
            res = [(fun, dim, iinst)
                        for (fun, dim), instances in self._data.items()
                            for iinst in instances
                                if not self.done((fun, dim, iinst))]
        return res
    def missing_trials(self, problem):
        """return number of missing trials for this problem instance

        compared to the other instances of the same problem or 1 when no
        trials were found at all.
        """
        return max((1, max([0] + self.trials(problem)))) - len(self.data_of_instance(problem))
    def initial_solution_proposal(self, problem, nonzero_odds=14):
        """return allzeros in the first of any ``nonzero_odds + 1`` trials

        if ``nonzero_odds > 0``, and otherwise
        ``problem.initial_solution_proposal(#trials_done + 1)``.
        """
        if nonzero_odds != int(nonzero_odds):
            raise ValueError("nonzero_odds={} must be an int".format(nonzero_odds))
        trials = sum(self.trials(problem))
        if nonzero_odds == 0:
            return problem.initial_solution_proposal(trials + 1)
        return problem.initial_solution_proposal(trials if trials % (nonzero_odds + 1) else 0)

class BatchScheduler:
    """Facilitate to run a benchmarking experiment in independent batches.

    The batch scheduler crucially assumes that in each batch the same
    problems are given *in the same order* when calling `is_in_batch`.

    Pseudo code example::

        batch_to_execute = 0  # set current batch to execute

        suite = cocoex.Suite('bbob', '', '')
        batcher = cocoex.BatchScheduler(4, batch_to_execute)
        for problem in suite:
            if not batcher.is_in_batch(problem):
                continue
            # ... run optimizer on problem ...
    
    needs to be run, in accordance with the first argument to
    `BatchScheduler`, four times overall (e.g., in parallel) with
    ``batch_to_execute in (0,1,2,3)`` to generate the full experimental
    data.

    Details: to get a more even time distribution over all batches, it
    seems advisable that the number of functions is not divisible by the
    number of batches. That is, 4 (or 6 or 8 or 12) batches is not likely
    to be ideal on the `'bbob'` testbed of 24 functions.
    """
    def __init__(self, number_of_batches, batch_to_execute):
        """distribute over `number_of_batches` batches and execute

        the batch with number `batch_to_execute` which must obey
        ``0 <= batch_to_execute < number_of_batches``.
        """
        self.params = {n: v for (n, v) in locals().items() if n != 'self'}
        self.first_problem = None
        self.current_problem = None
        self.current_batch = -1
        if self.params['number_of_batches'] == 1 and self.params['batch_to_execute'] in (0, 1, None):
            print("number_of_batches == 1, hence running the full suite")
            self.params['batch_to_execute'] = 0
        elif (self.params['batch_to_execute'] < 0 or
              self.params['batch_to_execute'] >= self.params['number_of_batches']):
            raise ValueError("batch_to_execute == {} must be >= 0 and < {} == number_of_batches."
                            "\n  The first argument is the number of batches (nob),"
                            "\n  the second argument needs to 'sweep' from 0...nob-1."
                            .format(self.params['batch_to_execute'],
                                    self.params['number_of_batches']))
    def is_in_batch(self, problem):
        """return `True` iff the batch number for `problem` equals `batch_to_execute`

        as given as constructor argument. Assumes that ``id_function`` and
        ``dimension`` are attributes of `problem`.

        The batch number for `problem` is attributed using
        ``(problem.id_function, problem.dimension)`` by order of
        appearance, in that the number is incremented whenever either
        ``id_function`` or ``dimension`` changes. Instances that follow
        suit with the same function ID and dimension belong to the same
        batch.
        """
        if self.first_problem is None:
            self.first_problem = (problem.id_function, problem.dimension)
        if self.current_problem != (problem.id_function, problem.dimension):
            if (problem.id_function, problem.dimension) == self.first_problem:
                self.current_batch = -1
            self.current_problem = (problem.id_function, problem.dimension)
            self.current_batch += 1
        return self.current_batch % self.params['number_of_batches'] == self.params['batch_to_execute']



# -*- mode: cython -*-
#cython: language_level=3, boundscheck=False, c_string_type=str, c_string_encoding=ascii

import sys
import numpy as np
cimport numpy as np

from .exceptions import InvalidProblemException, NoSuchProblemException, NoSuchSuiteException

np.import_array()

known_suite_names = ["bbob",
                     "bbob-biobj", "bbob-biobj-ext",
                     "bbob-boxed",
                     "bbob-constrained",
                     "bbob-largescale",
                     "bbob-mixint", "bbob-biobj-mixint",
                     "bbob-noisy",
                     ]
known_suites = known_suite_names  # provide the expected name too
_known_suite_names = ["bbob",
                      "bbob-biobj", "bbob-biobj-ext",
                      "bbob-boxed",
                      "bbob-constrained",
                      "bbob-constrained-active-only", "bbob-constrained-no-disguise",
                      "bbob-largescale",
                      "bbob-mixint", "bbob-biobj-mixint",
                      "bbob-noisy",
                      ]

__all__ = ['Observer', 'Problem', 'Suite', 'known_suite_names', 'known_suites']


cdef extern from "coco.h":
    ctypedef struct coco_problem_t:
        pass
    ctypedef struct coco_observer_t:
        pass
    ctypedef struct coco_suite_t:
        pass

    const char* coco_set_log_level(const char *level)

    coco_observer_t *coco_observer(const char *observer_name, const char *options)
    void coco_observer_free(coco_observer_t *self)
    coco_problem_t *coco_problem_add_observer(coco_problem_t *problem,
                                              coco_observer_t *observer)
    const char *coco_observer_get_result_folder(const coco_observer_t *observer)
    void coco_observer_signal_restart(coco_observer_t *observer, coco_problem_t *problem)

    coco_suite_t *coco_suite(const char *suite_name, const char *suite_instance,
                             const char *suite_options)
    void coco_suite_free(coco_suite_t *suite)
    void coco_problem_free(coco_problem_t *problem)

    void coco_problem_get_initial_solution(coco_problem_t *problem, double *x)
    void coco_evaluate_function(coco_problem_t *problem, const double *x, double *y)
    void coco_evaluate_constraint(coco_problem_t *problem, const double *x, double *y)
    void coco_recommend_solution(coco_problem_t *problem, const double *x)

    int coco_logger_biobj_feed_solution(coco_problem_t *problem, const size_t evaluation, const double *y)
    coco_problem_t *coco_suite_get_problem_by_function_dimension_instance(coco_suite_t *suite, const size_t function,
                                                                          const size_t dimension, const size_t instance)

    coco_problem_t* coco_suite_get_next_problem(coco_suite_t*, coco_observer_t*)
    coco_problem_t* coco_suite_get_problem(coco_suite_t *, const size_t)

    size_t coco_problem_get_suite_dep_index(const coco_problem_t* problem)
    size_t coco_problem_get_dimension(const coco_problem_t *problem)
    size_t coco_problem_get_number_of_objectives(const coco_problem_t *problem)
    size_t coco_problem_get_number_of_constraints(const coco_problem_t *problem)
    size_t coco_problem_get_number_of_integer_variables(const coco_problem_t *problem)
    const char *coco_problem_get_id(const coco_problem_t *problem)
    const char *coco_problem_get_name(const coco_problem_t *problem)
    const double *coco_problem_get_smallest_values_of_interest(const coco_problem_t *problem)
    const double *coco_problem_get_largest_values_of_interest(const coco_problem_t *problem)
    const double *coco_problem_get_largest_fvalues_of_interest(const coco_problem_t *problem)
    # double coco_problem_get_final_target_fvalue1(const coco_problem_t *problem)
    size_t coco_problem_get_evaluations(const coco_problem_t *problem)
    size_t coco_problem_get_evaluations_constraints(const coco_problem_t *problem)
    void coco_reset_seeds()
    double coco_problem_get_best_observed_fvalue1(const coco_problem_t *problem)
    int coco_problem_final_target_hit(const coco_problem_t *problem)
    void bbob_problem_best_parameter_print(const coco_problem_t *problem)
    void bbob_biobj_problem_best_parameter_print(const coco_problem_t *problem)


cdef bytes _bstring(s):
    if type(s) is bytes:
        return <bytes>s
    if isinstance(s, (str, unicode)):
        return s.encode('ascii')  # why not <bytes>s.encode('ascii') ?
    else:
        raise TypeError("expect a string, got %s" % str(type(s)))


cdef coco_observer_t* _current_observer


cdef class Suite:
    """see __init__.py"""
    cdef coco_suite_t* suite  # AKA _self
    cdef coco_problem_t* _current_problem
    cdef bytes _name  # used in @property name
    cdef bytes _instance
    cdef bytes _options
    cdef current_problem_  # name _current_problem is taken
    cdef _current_index
    cdef _ids
    cdef _indices
    cdef _names
    cdef _dimensions
    cdef _number_of_objectives
    cdef initialized

    def __cinit__(self, suite_name, suite_instance, suite_options):
        cdef np.npy_intp shape[1]  # probably completely useless
        self._name = _bstring(suite_name)
        self._instance = _bstring(suite_instance if suite_instance is not None else "")
        self._options = _bstring(suite_options if suite_options is not None else "")
        self._current_problem = NULL
        self.current_problem_ = None
        self._current_index = None
        self.initialized = False
        self._initialize()
        assert self.initialized

    cdef _initialize(self):
        """sweeps through `suite` to collect indices and id's to operate by
        direct access in the remainder"""
        cdef np.npy_intp shape[1]  # probably completely useless
        cdef coco_suite_t* suite
        cdef coco_problem_t* p
        cdef bytes _old_level
        coco_reset_seeds()
        if self.initialized:
            self.reset()
        self._ids = []
        self._indices = []
        self._names = []
        self._dimensions = []
        self._number_of_objectives = []

        try:
            suite = coco_suite(self._name, self._instance, self._options)
        except:
            raise NoSuchSuiteException(self._name)

        if suite == NULL:
            raise NoSuchSuiteException(self._name)

        while True:
            old_level = log_level('warning')
            p = coco_suite_get_next_problem(suite, NULL)
            log_level(old_level)
            if not p:
                break
            self._indices.append(coco_problem_get_suite_dep_index(p))
            self._ids.append(coco_problem_get_id(p))
            self._names.append(coco_problem_get_name(p))
            self._dimensions.append(coco_problem_get_dimension(p))
            self._number_of_objectives.append(coco_problem_get_number_of_objectives(p))
        coco_suite_free(suite)
        self.suite = coco_suite(self._name, self._instance, self._options)
        self.initialized = True
        return self
    def reset(self):
        """reset to original state, affecting `next_problem()`,
        `current_problem`, `current_index`"""
        self._current_index = None
        if self.current_problem_:
            self.current_problem_.free()
        self.current_problem_ = None
        self._current_problem = NULL
    def next_problem(self, observer=None):
        """`next_problem(observer=None)` returns the "next" problem in the
        `Suite`, on the first call or after `reset()` the first problem.

        `next_problem` serves to sweep through the `Suite` smoothly.
        """
        cdef size_t index
        global _current_observer
        if not self.initialized:
            raise ValueError("Suite has been finalized/free'ed")
        if self.current_problem_:
            self.current_problem_.free()
        if self._current_index is None:
            self._current_index = -1
        self._current_index += 1
        if self._current_index >= len(self):
            self._current_problem = NULL
            self.current_problem_ = None
            # self._current_index = -1  # or use reset?
        else:
            index = self.indices[self._current_index]  # "conversion" to size_t
            self._current_problem = coco_suite_get_problem(
                                        self.suite, index)
            self.current_problem_ = Problem_init(self._current_problem,
                                                True, self._name)
            self.current_problem_.observe_with(observer)
        return self.current_problem_
    def get_problem(self, id, observer=None):
        """`get_problem(self, id, observer=None)` returns a `Problem` instance,
        by default unobserved, using `id: str` or index (where `id: int`) to
        identify the desired problem.

        All values between zero and `len(self) - 1` are valid index values::

        >>> import cocoex as ex
        >>> suite = ex.Suite("bbob-biobj", "", "")
        >>> for index in range(len(suite)):
        ...     problem = suite.get_problem(index)
        ...     # work work work using problem
        ...     problem.free()

        A shortcut for `suite.get_problem(index)` is `suite[index]`, they are
        synonym.

        Details:
        - Here an `index` takes values between 0 and `len(self) - 1` and can in
          principle be different from the problem index in the benchmark suite.

        - This call does not affect the state of the `current_problem` and
          `current_index` attributes.

        - For some suites and/or observers, the `free()` method of the problem
          must be called before the next call of `get_problem`. Otherwise Python
          might just silently die, which is e.g. a known issue of the "bbob"
          observer.

        See also `ids`, `get_problem_by_function_dimension_instance`.
        """
        if not self.initialized:
            raise ValueError("Suite has been finalized/free'ed")
        index = id
        try:
            1 / (id == int(id))  # int(id) might raise an exception
        except:
            index = self._ids.index(id)
        try:
            return Problem_init(coco_suite_get_problem(self.suite, self._indices[index]),
                                True, self._name).observe_with(observer)
        except:
            raise NoSuchProblemException(self.name, str(id))

    def get_problem_by_function_dimension_instance(self, function, dimension, instance, observer=None):
        """returns a `Problem` instance, by default unobserved, using function,
        dimension and instance to identify the desired problem.

        If a suite contains multiple problems with the same function, dimension
        and instance, the first corresponding problem is returned.

        >>> import cocoex as ex
        >>> suite = ex.Suite("bbob-biobj", "", "")
        >>> problem = suite.get_problem_by_function_dimension_instance(1, 2, 3)
        >>> # work work work using problem
        >>> problem.free()

        Details:
        - Function, dimension and instance are integer values from 1 on.

        - This call does not affect the state of the `current_problem` and
          `current_index` attributes.

        - For some suites and/or observers, the `free()` method of the problem
          must be called before the next call of
          `get_problem_by_function_dimension_instance`. Otherwise Python might
          just silently die, which is e.g. a known issue of the "bbob" observer.
        """
        cdef size_t _function = function # "conversion" to size_t
        cdef size_t _dimension = dimension # "conversion" to size_t
        cdef size_t _instance = instance # "conversion" to size_t

        if not self.initialized:
            raise ValueError("Suite has been finalized/free'ed")
        try:
            return Problem_init(coco_suite_get_problem_by_function_dimension_instance(self.suite, _function,
                                                                                      _dimension, _instance),
                                True, self._name).observe_with(observer)
        except:
            raise NoSuchProblemException(self.name, 'function: {}, dimension: {}, instance: {}'.format(function,
                                                                                                       dimension,
                                                                                                       instance))

    def __getitem__(self, key):
        """`self[i]` is a synonym for `self.get_problem(i)`, see `get_problem`
        """
        return self.get_problem(key)

    def free(self):
        """free underlying C structures"""
        if self.suite:  # for some reason __dealloc__ cannot be called here
            coco_suite_free(self.suite)
        self.suite = NULL
        self.initialized = False  # not (yet) visible from outside
    def __dealloc__(self):
        if self.suite:
            coco_suite_free(self.suite)

    def find_problem_ids(self, *args, **kwargs):
        """has been renamed to `ids`"""
        raise NotImplementedError(
            "`find_problem_ids()` has been renamed to `ids()`")


    def ids(self, *id_snippets, get_problem=False, verbose=False):
        """`ids(*id_snippets, get_problem=False, verbose=False)`
        return all problem IDs that contain all of the `id_snippets`.

        An ID can be used for indexing, that is, when calling the method
        `get_problem(id)`.

        If `get_problem is True`, the problem for the first matching ID is
        returned.

        >>> import cocoex as ex
        >>> s = ex.Suite("bbob", "", "")
        >>> s.ids("f001", "d10", "i01")
        ['bbob_f001_i01_d10']

        We can sweep through all instances of the ellipsoidal function f10
        in 20-D of the BBOB suite like this::

        >>> import cocoex as ex
        >>> suite = ex.Suite("bbob", "", "")
        >>> ids = suite.ids("f010", "d20")
        >>> used_indices = []
        >>> for p in suite:
        ...     if p.id in ids:
        ...         # work work work with problem `p`
        ...         used_indices.append(p.index)
        >>> print(used_indices)
        [1575, 1576, 1577, 1578, 1579, 1580, 1581, 1582, 1583, 1584, 1585, 1586, 1587, 1588, 1589]

        A desired problem can also be filtered out during creation::

        >>> import cocoex as ex
        >>> f9 = ex.Suite("bbob", "",
        ...               "function_indices:9 dimensions:20 instance_indices:1-5")[0]
        >>> print(f9.id)
        bbob_f009_i01_d20

        """
        res = []
        for idx, id in enumerate(self._ids):
            if all([id.find(i) >= 0 for i in id_snippets]):
                if verbose:
                    print("  id=%s, index=%d" % (id, idx))
                res.append(id)
        if get_problem:
            return self.get_problem(res[0])
        return res

    @property
    def current_problem(self):
        """current "open/active" problem to be benchmarked"""
        return self.current_problem_
    @property
    def current_index(self):
        """index in the enumerator of all problems in this suite.

        Details: To get the index in the underlying C implementation, which
        usually matches `current_index` one-to-one, use::

        >>> import cocoex as ex
        >>> suite = ex.Suite("bbob", "", "")
        >>> suite.current_index is None
        True
        >>> suite.next_problem().id[-17:].lower()
        'bbob_f001_i01_d02'
        >>> suite.current_index, suite.indices[suite.current_index]
        (0, 0)

        """
        return self._current_index
    @property
    def problem_names(self):
        """list of problem names in this `Suite`, see also `ids`"""
        return list(self._names)
    @property
    def dimensions(self):
        """list of problem dimensions occurring at least once in this `Suite`"""
        return sorted(set(self._dimensions))
    @property
    def number_of_objectives(self):
        """list of number of objectives occurring in this `Suite`"""
        return sorted(set(self._number_of_objectives))
    @property
    def indices(self):
        """list of all problem indices, deprecated.

        These values are (only) used to call the underlying C structures.
        Indices used in the Python interface run between 0 and `len(self)`.
        """
        return list(self._indices)
    @property
    def name(self):
        """see __init__.py"""
        return self._name
    @property
    def instance(self):
        """instance of this suite as used to instantiate the suite via
        `Suite(name, instance, ...)`"""
        return self._instance
    @property
    def options(self):
        """options for this suite as used to instantiate the suite via
        `Suite(name, instance, options)`"""
        return self._options

    @property
    def info(self):
        return str(self)
    def __repr__(self):
        return 'Suite(%r, %r, %r)'  % (self.name, self.instance, self.options)  # angled brackets
    def __str__(self):
        return 'Suite("%s", "%s", "%s") with %d problem%s in dimension%s %s' \
            % (self.name, self.instance, self.options,
               len(self), '' if len(self) == 1 else 's',
               '' if len(self.dimensions) == 1 else 's',
               '%d=%d' % (min(self.dimensions), max(self.dimensions)))
    def __len__(self):
        return len(self._indices)

    def __iter__(self):
        """iterator over self.

        CAVEAT: this function uses `next_problem` and has a side effect on the
        state of the `current_problem` and `current_index` attributes. `reset()`
        rewinds the suite to the initial state. """
        if 1 < 3:
            s = self
            s.reset()
        else:
            s = Suite(self.name, self.instance, self.options)
        try:
            while True:
                try:
                    problem = s.next_problem()
                    if problem is None:
                        return  # StopIteration is deprecated
                        # raise StopIteration
                except NoSuchProblemException:
                    return  # StopIteration is deprecated
                    # raise StopIteration
                yield problem
        except:
            raise
        finally:  # makes this ctrl-c safe, at least it should
            s is self or s.free()


cdef class Observer:
    """see __init__.py"""
    cdef coco_observer_t* _observer
    cdef bytes _name
    cdef bytes _options
    cdef _state

    def __cinit__(self, name, options):
        if isinstance(options, dict):
            s = str(options).replace(',', ' ')
            for c in ["u'", 'u"', "'", '"', "{", "}"]:
                s = s.replace(c, '')
            options = s
        self._name = _bstring(name)
        self._options = _bstring(options if options is not None else "")
        self._observer = coco_observer(self._name, self._options)
        self._state = 'initialized'

    def _update_current_observer_global(self):
        """assign the global _current_observer variable to self._observer,
        for purely technical reasons"""
        global _current_observer
        _current_observer = self._observer

    def observe(self, problem):
        """`observe(problem)` let `self` observe the `problem: Problem` by
        calling `problem.observe_with(self)`.
        """
        problem.observe_with(self)
        return self

    def signal_restart(self, problem: Problem):
        """Signal a restart on `problem: Problem` by calling `coco_observer_signal_restart`.
        """
        coco_observer_signal_restart(self._observer, problem.problem)

    @property
    def name(self):
        """name of the observer as used with `Observer(name, ...)` to instantiate
        `self` before.
        """
        return self._name
    @property
    def options(self):
        return self._options
    @property
    def state(self):
        return self._state
    @property
    def result_folder(self):
        return coco_observer_get_result_folder(self._observer)

    def free(self):
        self.__dealloc__()
        self._observer = NULL
        self._state = 'deactivated'
    def __dealloc__(self):
        if self._observer !=  NULL:
            coco_observer_free(self._observer)


cdef Problem_init(coco_problem_t* problem, free=True, suite_name=None):
    """`Problem` class instance initialization wrapper passing
    a `problem_t*` C-variable to `__init__`.

    This is necessary because __cinit__ cannot be defined as cdef, only as def.
    """
    res = Problem()
    res._suite_name = suite_name
    return res._initialize(problem, free)


cdef class Problem:
    """see __init__.py"""
    cdef coco_problem_t* problem
    cdef np.ndarray y_values  # argument for coco_evaluate
    cdef np.ndarray constraint_values  # argument for coco_evaluate
    cdef np.ndarray x_initial  # argument for coco_problem_get_initial_solution
    cdef np.ndarray _lower_bounds
    cdef np.ndarray _upper_bounds
    cdef np.ndarray _largest_fvalues_of_interest
    cdef size_t _number_of_variables
    cdef size_t _number_of_objectives
    cdef size_t _number_of_constraints
    cdef size_t _number_of_integer_variables
    cdef _suite_name  # for the record
    cdef _list_of_observers  # for the record
    cdef _problem_index  # for the record, this is not public but used in index property
    cdef _do_free
    cdef _initial_solution_proposal_calls
    cdef initialized
    def __cinit__(self):
        cdef np.npy_intp shape[1]
        self.initialized = False  # all done in _initialize
    cdef _initialize(self, coco_problem_t* problem, free=True):
        cdef np.npy_intp shape[1]
        if self.initialized:
            raise RuntimeError("Problem already initialized")
        if problem == NULL:
            raise ValueError("in Problem._initialize(problem,...): problem is NULL")
        self.problem = problem
        self._problem_index = coco_problem_get_suite_dep_index(self.problem)
        self._do_free = free
        self._list_of_observers = []
        # _problem_suite = _bstring(problem_suite)
        # self.problem_suite = _problem_suite
        # Implicit type conversion via passing safe,
        # see http://docs.cython.org/src/userguide/language_basics.html
        self._number_of_variables = coco_problem_get_dimension(self.problem)
        self._number_of_objectives = coco_problem_get_number_of_objectives(self.problem)
        self._number_of_constraints = coco_problem_get_number_of_constraints(self.problem)
        self._number_of_integer_variables = coco_problem_get_number_of_integer_variables(self.problem)
        self.y_values = np.zeros(self._number_of_objectives)
        self.constraint_values = np.zeros(self._number_of_constraints)
        self.x_initial = np.zeros(self._number_of_variables)
        self._initial_solution_proposal_calls = 0
        ## FIXME: Inefficient because we copy the bounds instead of
        ## sharing the data.
        self._lower_bounds = -np.inf * np.ones(self._number_of_variables)
        self._upper_bounds = np.inf * np.ones(self._number_of_variables)
        # self.test_bounds = coco_problem_get_smallest_values_of_interest(self.problem)  # fails
        for i in range(self._number_of_variables):
            if coco_problem_get_smallest_values_of_interest(self.problem) is not NULL:
                self._lower_bounds[i] = coco_problem_get_smallest_values_of_interest(self.problem)[i]
            if coco_problem_get_largest_values_of_interest(self.problem) is not NULL:
                self._upper_bounds[i] = coco_problem_get_largest_values_of_interest(self.problem)[i]
        self._largest_fvalues_of_interest = None
        self.initialized = True
        return self

    def constraint(self, x):
        """see __init__.py"""
        if self.number_of_constraints <= 0:
            return  # return None, prevent Python kernel from dying
            # or should we return `[]` for zero constraints?
            # `[]` is more likely to produce quietly unexpected result?
        cdef np.ndarray[double, ndim=1, mode="c"] _x
        x = np.asarray(x, dtype=np.double, order='C')
        if np.size(x) != self.number_of_variables:
            raise ValueError(
                "Dimension, `np.size(x)==%d`, of input `x` does " % np.size(x) +
                "not match the problem dimension `number_of_variables==%d`."
                             % self.number_of_variables)
        _x = x  # this is the final type conversion
        if self.problem is NULL:
            raise InvalidProblemException()
        coco_evaluate_constraint(self.problem,
                               <double *>np.PyArray_DATA(_x),
                               <double *>np.PyArray_DATA(self.constraint_values))
        return np.array(self.constraint_values, copy=True)
    def recommend(self, arx):
        """Recommend a solution, return `None`.

        The recommendation replaces the last evaluation or recommendation
        for the assessment of the algorithm.
        """
        cdef np.ndarray[double, ndim=1, mode="c"] _x
        x = np.asarray(arx, dtype=np.double, order='C')
        if np.size(x) != self.number_of_variables:
            raise ValueError(
                "Dimension, `np.size(x)==%d`, of input `x` does " % np.size(x) +
                "not match the problem dimension `number_of_variables==%d`."
                % self.number_of_variables)
        _x = x  # this is the final type conversion
        if self.problem is NULL:
            raise InvalidProblemException()
        coco_recommend_solution(self.problem, <double *>np.PyArray_DATA(_x))

    def logger_biobj_feed_solution(self, evaluation, y):
        """Feed the given solution to logger_biobj in order to reconstruct its
        output.

        Return 1 if the given solution updated the archive and 0 otherwise.

        Used by preprocessing when updating the .info, .dat and .tdat files
        with new indicator reference values.
        """
        cdef size_t _evaluation = evaluation # "conversion" to size_t
        cdef np.ndarray[double, ndim=1, mode="c"] _y
        y = np.asarray(y, dtype=np.double, order='C')
        if np.size(y) != self.number_of_objectives:
            raise ValueError(
                "Dimension, `np.size(y)==%d`, of input `y` does " % np.size(y) +
                "not match the number of objectives `number_of_objectives==%d`."
                             % self.number_of_objectives)
        _y = y  # this is the final type conversion
        if self.problem is NULL:
            raise InvalidProblemException()
        return coco_logger_biobj_feed_solution(self.problem, _evaluation, <double *>np.PyArray_DATA(_y))


    def add_observer(self, observer):
        """`add_observer(self, observer: Observer)`, see `observe_with`.
        """
        return self.observe_with(observer)

    def observe_with(self, observer):
        """`observe_with(self, observer: Observer)` attaches an observer
        to this problem.

        Attaching an observer can be considered as wrapping the observer
        around the problem. For the observer to be finalized, the problem
        must be free'd (implictly or explicitly).

        Details: `observer` can be `None`, in which case nothing is done.

        See also: class `Observer`
        """
        if observer:
            assert self.problem
            observer._update_current_observer_global()
            self.problem = coco_problem_add_observer(self.problem, _current_observer)
            self._list_of_observers.append(observer)
        return self

    def _f0(self, x):
        """"inofficial" interface to `self` with target f-value of zero. """
        return self(x) - self.final_target_fvalue1

    def initial_solution_proposal(self, restart_number=None):
        """return initial solution proposals.

        The proposal is different for each consecutive call without
        argument and for each `restart_number` and may be different under
        repeated calls with the same `restart_number`.
        ``self.initial_solution_proposal(0)`` is the same as
        ``self.initial_solution`` and is always feasible.

        Conceptual example::

            # given: a suite instance, a budget, and fmin
            for problem in suite:
                # restart until budget is (over-)exhausted
                while problem.evaluations < budget and not problem.final_target_hit:
                    fmin(problem, problem.initial_solution_proposal())

        Details: by default, the first proposal is the domain middle or the
        (only) known feasible solution. Subsequent proposals are
        coordinate-wise sampled uniformly at random for discrete variables
        and otherwise as the sum of two iid uniformly distributed random
        variates, scaled to have positive density only within the domain
        boundaries for unconstrained problems and around the feasible
        initial solution +-1 for constrained problems.

        On the ``'bbob'`` suite their density
        is 0.2 * (x / 5 + 1) for x in [-5, 0] and
        0.2 * (1 - x / 5) for x in [0, 5] and zero otherwise.

        """
        if restart_number is None:
            restart_number = self._initial_solution_proposal_calls
            self._initial_solution_proposal_calls += 1  # count calls without explicit argument
        if restart_number <= 0:
            return self.initial_solution
        rv_triangular = np.random.rand(self.dimension) + np.random.rand(self.dimension)  # in [0, 2]
        for i in range(self.number_of_integer_variables):
            rv_triangular[i] = np.random.randint(self.lower_bounds[i], self.upper_bounds[i] + 1)
        if self.number_of_constraints > 0:
            # returns self.initial_solution + rv_triangular - 1
            rv_triangular[self._number_of_integer_variables:] += -1 + self.initial_solution[self._number_of_integer_variables:]
        else:
            # returns lb + rv * (ub - lb) / 2
            rv_triangular[self._number_of_integer_variables:] *= (self.upper_bounds[self._number_of_integer_variables:] -
                                                                  self.lower_bounds[self._number_of_integer_variables:]) / 2
            rv_triangular[self._number_of_integer_variables:] += self.lower_bounds[self._number_of_integer_variables:]
        return rv_triangular
    @property
    def initial_solution(self):
        """return feasible initial solution"""
        coco_problem_get_initial_solution(self.problem,
                                          <double *>np.PyArray_DATA(self.x_initial))
        return np.array(self.x_initial, copy=True)
    @property
    def observers(self):
        """list of observers wrapped around this problem"""
        return self._list_of_observers
    @property
    def is_observed(self):
        """problem ``p`` is observed ``p.is_observed`` times.

        See also: the list of observers in property `observers`.
        """
        return len(self._list_of_observers)

    property number_of_variables:  # this is cython syntax, not known in Python
        # this is a class definition which is instantiated automatically!?
        """Number of variables this problem instance expects as input."""
        def __get__(self):
            return self._number_of_variables
    @property
    def dimension(self):
        """alias for `number_of_variables` of the input space"""
        return self._number_of_variables
    @property
    def number_of_objectives(self):
        "number of objectives, if equal to 1, call returns a scalar"
        return self._number_of_objectives
    @property
    def number_of_constraints(self):
        "number of constraints"
        return self._number_of_constraints
    @property
    def number_of_integer_variables(self):
        "number of integer variables"
        return self._number_of_integer_variables
    @property
    def lower_bounds(self):
        """depending on the test bed, these are not necessarily strict bounds
        """
        return self._lower_bounds
    @property
    def upper_bounds(self):
        """depending on the test bed, these are not necessarily strict bounds
        """
        return self._upper_bounds
    @property
    def evaluations(self):
        return coco_problem_get_evaluations(self.problem)
    @property
    def evaluations_constraints(self):
        return coco_problem_get_evaluations_constraints(self.problem)
    @property
    def final_target_hit(self):
        """return `True` if the final target has been hit, `False` otherwise
        """
        assert(self.problem)
        return bool(coco_problem_final_target_hit(self.problem))
    #@property
    #def final_target_fvalue1(self):
    #    assert(self.problem)
    #    return coco_problem_get_final_target_fvalue1(self.problem)
    @property
    def best_observed_fvalue1(self):
        assert(self.problem)
        return coco_problem_get_best_observed_fvalue1(self.problem)
    @property
    def largest_fvalues_of_interest(self):
        "largest f-values of interest (defined only for multi-objective problems)"
        assert(self.problem)
        if self._number_of_objectives > 1 and coco_problem_get_largest_fvalues_of_interest(self.problem) is not NULL:
            self._largest_fvalues_of_interest = np.asarray(
                [coco_problem_get_largest_fvalues_of_interest(self.problem)[i] for i in range(self._number_of_objectives)])
        return self._largest_fvalues_of_interest

    def _best_parameter(self, what=None):
        if what == 'print':
            if self._number_of_objectives == 2:
                bbob_biobj_problem_best_parameter_print(self.problem)
            else:
                bbob_problem_best_parameter_print(self.problem)

    def free(self, force=False):
        """Free the given test problem.

        Not strictly necessary (unless, possibly, for the observer). `free`
        ensures that all files associated with the problem are closed as
        soon as possible and any memory is freed. After free()ing the
        problem, all other operations are invalid and will raise an
        exception.
        """
        if self.problem != NULL and (self._do_free or force):
            coco_problem_free(self.problem)
            self.problem = NULL

    def __dealloc__(self):
        # see http://docs.cython.org/src/userguide/special_methods.html
        # free let the problem_free() call(s) in coco_suite_t crash, hence
        # the possibility to set _do_free = False
        if self._do_free and self.problem != NULL:  # this is not guaranteed to work, see above link
            coco_problem_free(self.problem)

    # def __call__(self, np.ndarray[double, ndim=1, mode="c"] x):
    def __call__(self, x):
        """return objective function value of input `x`"""
        cdef np.ndarray[double, ndim=1, mode="c"] _x
        assert self.initialized
        x = np.asarray(x, dtype=np.double, order='C')
        if np.size(x) != self.number_of_variables:
            raise ValueError(
                "Dimension, `np.size(x)==%d`, of input `x` does " % np.size(x) +
                "not match the problem dimension `number_of_variables==%d`."
                             % self.number_of_variables)
        _x = x  # this is the final type conversion
        if self.problem is NULL:
            raise InvalidProblemException()
        coco_evaluate_function(self.problem,
                               <double *>np.PyArray_DATA(_x),
                               <double *>np.PyArray_DATA(self.y_values))
        if self._number_of_objectives == 1:
            return self.y_values[0]
        return np.array(self.y_values, copy=True)

    @property
    def id(self):
        "id as string without spaces or weird characters"
        if self.problem is not NULL:
            return coco_problem_get_id(self.problem)

    def _parse_id(self, substr):
        "search `substr` in `id` and return converted `int` up to '_'"
        if self.problem is NULL:
            return None
        i = self.id.find(substr)
        if i < 0:
            raise ValueError()
        return int(self.id[i + len(substr):].split('_')[0])

    @property
    def id_function(self):
        "see __init__.py"
        try:
            return self._parse_id('_f')
        except ValueError:
            raise ValueError("cannot deduce function id from '%s'" % self.id)

    @property
    def id_triple(self):
        "see __init__.py"
        return self.id_function, self.dimension, self.id_instance

    @property
    def id_instance(self):
        "see __init__.py"
        try:
            return self._parse_id('_i')
        except ValueError:
            raise ValueError("cannot deduce instance id from '%s'" % self.id)

    @property
    def name(self):
        if self.problem is not NULL:
            return coco_problem_get_name(self.problem)

    @property
    def index(self):
        """problem index in the benchmark `Suite` of origin"""
        return self._problem_index

    @property
    def suite(self):
        """benchmark suite this problem is from"""
        return self._suite_name

    @property
    def info(self):
        """see __init__.py"""
        return str(self)

    def __str__(self):
        if self.problem is not NULL:
            dimensional = "%d-dimensional" % self.dimension
            objective = "%s-objective" % {
                    1: 'single',
                    2: 'bi'}.get(self.number_of_objectives,
                                 str(self.number_of_objectives))
            constraints = "" if self.number_of_constraints == 0 else (
                " with %d constraint%s" % (self.number_of_constraints,
                                           "s" if self.number_of_constraints > 1 else "")
                )
            integer_variables = "" if self.number_of_integer_variables == 0 else (
                " %s %d integer variable%s" % ("and" if constraints != "" else "with",
                self.number_of_integer_variables,
                "s" if self.number_of_integer_variables > 1 else "")
            )
            return '%s: a %s %s problem%s%s (problem %d of suite "%s" with name "%s")' % (
                    self.id, dimensional, objective, constraints, integer_variables, self.index,
                    self.suite, self.name)
                    # self.name.replace(self.name.split()[0],
                    #               self.name.split()[0] + "(%d)"
                    #               % (self.index if self.index is not None else -2)))
        else:
            return "finalized/invalid problem"

    def __repr__(self):
        if self.problem is not NULL:
            return "<%s(), id=%r>" % (
                    repr(self.__class__).split()[1][1:-2],
                    # self.problem_suite, self.problem_index,
                    self.id)
        else:
            return "<finalized/invalid problem>"

    def __enter__(self):
        """Allows ``with Suite(...)[index] as problem:`` (or ``Suite(...).get_problem(...)``)
        """
        return self
    def __exit__(self, exception_type, exception_value, traceback):
        try:
            self.free()
        except:
            pass


def log_level(level=None):
    """`log_level(level=None)` return current log level and
    set new log level if `level is not None and level`.

    `level` must be 'error' or 'warning' or 'info' or 'debug', listed
    with increasing verbosity, or '' which doesn't change anything.
    """
    cdef bytes _level = _bstring(level if level is not None else "")
    return coco_set_log_level(_level)



import numpy as np
from setuptools import Extension, setup


extensions = []
extensions.append(Extension(name="cocoex.interface",
                            sources=["src/cocoex/coco.c", "src/cocoex/interface.pyx"],
                            define_macros=[("NPY_NO_DEPRECATED_API", "NPY_1_7_API_VERSION")],
                            include_dirs=[np.get_include()]))
extensions.append(Extension(name="cocoex.bare_problem",
                            sources=["src/cocoex/coco.c", "src/cocoex/bare_problem.pyx"],
                            define_macros=[("NPY_NO_DEPRECATED_API", "NPY_1_7_API_VERSION")],
                            include_dirs=[np.get_include()]))

setup(ext_modules=extensions)
